@@ -8,7 +8,7 @@ use Carp qw/croak/;
 use constant PIPE_READ_SIZE => 1024;
 
 use vars qw( $VERSION );
-$VERSION = '1.10';
+$VERSION = '1.7';
 
 BEGIN {
     use Exporter ();
@@ -29,7 +29,7 @@ A basic example:
  use Spawn::Safe;
  use Data::Dumper;
  my $results = spawn_safe({ argv => [qw{ ls -al /var/ }], timeout => 2 });
- die Dumper $results;
+ die Dumper $results
 
 As a replacement for backticks:
 
@@ -115,18 +115,17 @@ binary's stdout or stderr will also be made available, but since the binary was
 forcefully terminated, the data may be incomplete.
 
 =item * If the binary was executed successfully and ran to completion, the keys
-'exit_code', 'stdout, and 'stderr', will all be available.
+'exit_code, 'stdout, and 'stderr', will all be available.
 
 =back
 
 If passed invalid parameters, spawn_safe will croak.
 
-Please note that when specifying a timeout, alarm() is no longer used. If the
-clock is stepped significantly backwards during a timeout, a possibly false
-timeout error may be thrown. Timeout accuracy should be within one second.
-
-If a timeout does occur, the spawned program will be sent a SIGKILL before
-spawn_safe returns.
+Please note that when specifying a timeout, alarm() will be used in order to
+implement the timeout. If the caller has an active SIGALRM handler, it will be
+overwritten during the call to spawn_safe and then restored afterwards. Any
+pending alarms should be cleared before calling spawn_safe regardless of whether
+or not the there is an active SIGALRM handler.
 
 =cut
 
@@ -134,7 +133,6 @@ sub spawn_safe {
     my ( $params ) = @_;
     my @binary_and_params;
     my $timeout = undef;
-    my $start_time;
     my $new_env;
 
     if ( ref $params eq '' ) {
@@ -165,8 +163,8 @@ sub spawn_safe {
     my ( $parent_read_stderr, $child_write_stderr );
     my ( $parent_signal,      $child_wait );
     my ( $parent_read_errors, $child_write_errors );
-
-    my ( $read_stdout, $read_stderr, $read_errors ) = ( '' ) x 3;
+    my ( $read_stdout,        $read_stderr, $read_errors ) = ( '' ) x 3;
+    $timeout ||= undef;
 
     pipe( $parent_read_stdout, $child_write_stdout ) || die $!;
     pipe( $parent_read_stderr, $child_write_stderr ) || die $!;
@@ -208,26 +206,28 @@ sub spawn_safe {
         exit 42;
     }
 
-    close( $child_write_stdout );
-    close( $child_write_stderr );
-    close( $child_wait );
-    close( $child_write_errors );
-    my $sel = IO::Select->new( $parent_read_stdout, $parent_read_stderr, $parent_read_errors )
-        || die "Failed to create IO::Select object!";
-    close( $parent_signal );
+    eval {
+        close( $child_write_stdout );
+        close( $child_write_stderr );
+        close( $child_wait );
+        close( $child_write_errors );
+        # Looks silly, but we have to localize $SIG{'ALRM'} AND set it back to
+        # what it was when we were called, or else local will reset it to the
+        # default handler. 
+        local $SIG{'ALRM'} = $SIG{'ALRM'} || 'DEFAULT';
+        my $sel = IO::Select->new( $parent_read_stdout, $parent_read_stderr, $parent_read_errors );
+        close( $parent_signal );
+        if ( $timeout ) {
+            $SIG{'ALRM'} = sub { die "a\n" };    # \n required
+            alarm( $timeout );
+        }
 
-    # Don't bother calling time if we're never going to timeout.
-    $start_time = defined $timeout ? time() : 1;
-    my $select_time = $timeout;
-    MAIN_WHILE: while ( 1 ) {
-        foreach my $readme ( $sel->can_read( $select_time ) ) {
-            # Won't be set if there was a timeout.
-            if ( $readme ) {
+        while ( $sel->count() > 0 ) {
+            foreach my $readme ( $sel->can_read( undef ) ) {
                 my $read;
                 my $r = sysread( $readme, $read, PIPE_READ_SIZE );
-                if ( ( !defined $r ) || ( $r < 1 ) ) {
+                if ( $r < 1 ) {
                     $sel->remove( $readme );
-                    if ( $sel->count() == 0 ) { last MAIN_WHILE; }
                 } elsif ( $readme == $parent_read_stdout ) {
                     $read_stdout .= $read;
                 } elsif ( $readme == $parent_read_stderr ) {
@@ -239,47 +239,37 @@ sub spawn_safe {
                 }
             }
         }
-        if ( defined $timeout ) {
-            # We do a little gymnastics here to check if the time has rolled
-            # backwards (ie, ntpd stepped the time backwards). If it went
-            # backwards, there's no way to tell how long we've waited, so
-            # it's probably safer to assume we've waited too long. Hopefully
-            # steps backwards will be infrequent, as ntpd usually slews rather
-            # than steps.
-            # If the time rolls over, we should end up with a hugely negative
-            # $timeout after subtraction, so that will probably trigger a
-            # timeout as well. Imperfect, but somewhat better than waiting
-            # forever. Fortunately this probably won't ever come up.
-            my $timenow = time();
-            $select_time = $timeout - ( $timenow - $start_time );
-            if ( $timenow < $start_time || $select_time <= 0 ) {
-                undef $start_time;
-                last;
-            }
-        }
-    }
-    # Did we timeout? undef $start_time is our timeout flag.
-    if ( defined $start_time ) {
         waitpid( $child_pid, 0 );
-        $exit_code = $? >> 8;
-    }
+        if ( $timeout ) {
+            alarm( 0 );
+        }
+    };
 
+    $exit_code = $? >> 8;
     close( $parent_read_stdout );
     close( $parent_read_stderr );
     close( $parent_read_errors );
 
-    if ( !defined $start_time ) {
+    # Errors from eval?
+    if ( $@ ) {
         # If the child is still running, kill it.
         if ( waitpid( $child_pid, WNOHANG ) != -1 ) {
             kill( 9, $child_pid );
             waitpid( $child_pid, 0 );
         }
+        if ( !defined $exit_code ) {
+            $exit_code = $? >> 8;
+        }
 
-        return {
-            'error' => 'timed out',
-            'stdout' => $read_stdout,
-            'stderr' => $read_stderr,
-        };
+        if ( $@ eq "a\n" ) {
+            return {
+                'error' => 'timed out',
+                'stdout' => $read_stdout,
+                'stderr' => $read_stderr,
+            };
+        } else {
+            die 'Should not be here!';
+        }
     }
 
     if ( $read_errors ) {
@@ -294,21 +284,6 @@ sub spawn_safe {
 
 =head1 CHANGES
 
-=head2 Version 1.10 - 2011-05-11, jeagle
-
-Correct timeout handling. Attempt to correct unit tests for MSWin32, but
-there seems to be an issue with IO::Select preventing it from working
-properly.
-
-=head2 Version 1.9 - 2011-05-10, jeagle
-
-Don't use clock_gettime(), use time() and return a timeout if time steps
-backwards.
-
-=head2 Version 1.8 - 2011-05-09, jeagle
-
-Clean up docs, stop using SIGALARM for timeouts.
-
 =head2 Version 1.7 - 2010-07-09, jeagle
 
 Clean up for release to CPAN.
